﻿//-----------------------------------------------------------------------
// <copyright file="BlockCreator.cs" company="Microsoft">
//     Copyright (c) Microsoft. All rights reserved.
// </copyright>
// -----------------------------------------------------------------------

namespace BtmToTrfmConvertor
{
    using System;
    using Microsoft.ApplicationServer.Integration.Transforms.MapOM;
    using Logger;

    /// <summary>
    /// Does the changes in a Block
    /// </summary>
    class BlockCreator
    {
        // Adds statments and initialisations to a block
        internal static void AddToBlock(ref Block block, object item)
        {
            if (item is Assignment)
            {
                block.Statements.Add(item as Assignment);
            }
            else if (item is VariableInitialization)
            {
                VariableInitialization variable = item as VariableInitialization;
                if (variable.Initializer is LogicalExpression)
                {
                    LogicalExpression expression = variable.Initializer as LogicalExpression;
                    Optimiser.SetType(ref expression);
                    (variable).Initializer = expression;
                }

                block.Statements.Add(variable);
            }
            else if (item is Mapeach)
            {
                block.Statements.Add(item as Mapeach);
            }
            else if (item is Foreach)
            {
                block.Statements.Add(item as Foreach);
            }
            else if (item is ConditionalAssignment)
            {
                block.Statements.Add(item as ConditionalAssignment);
            }
            else if (item is ListInitialization)
            {
                block.Statements.Add(item as ListInitialization);
            }
            else if (item is AddItemToList)
            {
                block.Statements.Add(item as AddItemToList);
            }
            else
            {
                Controller.log.debug(MsgType.UNSUCCESSFUL_CONVERSION, "an object of type " + item.GetType().Name + "failed to get added to the current scope");
            }
        }
       
        // the function checks if the obj has atleast one source path in the curent context
        internal static bool BelongsToCurrentBlock(object obj, string loopVariable)
        {
            string sourcePath = string.Empty;
            if (obj is Assignment)
            {
                return true; // assignment statements are basically target links and always belong to the current scope
            }
            else if (obj is VariableInitialization)
            {
                VariableInitialization variable = obj as VariableInitialization;
                if (variable.Initializer is FunctionInvocationExpression)
                {
                    return BlockCreator.FunctoidInCurrentBlock(variable.Initializer as FunctionInvocationExpression, loopVariable);
                }
                else if (variable.Initializer is ContextExpression)
                {
                    return true; // getCurrentItem or getCurrentPosition which must belong to current scope anyway
                }
                else if (variable.Initializer is ScriptInvocationExpression)
                {
                    return BlockCreator.FunctoidInCurrentBlock(variable.Initializer as ScriptInvocationExpression, loopVariable);
                }
                else
                {
                    bool hasInputPaths = false;
                    TextBasedExpression textBasedExpression = variable.Initializer as TextBasedExpression;
                    foreach (NameBinding nameBinding in textBasedExpression.Aliases)
                    {
                        if (nameBinding.Expression is PathExpression)
                        {
                            PathExpression pathExpression = nameBinding.Expression as PathExpression;
                            sourcePath = pathExpression.Path;
                            if (BlockCreator.CompareSourceRootPaths(sourcePath, loopVariable))
                            {
                                return true;
                            }
                            else
                            {
                                hasInputPaths = true;
                            }
                        }
                    }

                    // returns true if it has no inputs. So must be a converted script functoid
                    return !hasInputPaths;
                }
            }
            else if (obj is Mapeach)
            {
                sourcePath = ((Mapeach)obj).SourceLoopPath.Path;
            }
            else if (obj is Foreach)
            {
                sourcePath = ((Foreach)obj).SourceLoopPath.Path;
            }
            else if (obj is ConditionalAssignment)
            {
                ConditionalAssignment assignment = obj as ConditionalAssignment;
                if (assignment.From is ConstantExpression)
                {
                    return true;
                }
                else
                {
                    PathExpression pathExpression = assignment.From as PathExpression;
                    sourcePath = pathExpression.Path;
                }
            }
            else
            {
                return true;
            }

            return BlockCreator.CompareSourceRootPaths(sourcePath, loopVariable);
        }

        private static bool FunctoidInCurrentBlock(ScriptInvocationExpression invocationExpression, string loopVariable)
        {
            string sourcePath = string.Empty;
            bool hasPathExpressions = false;

            // checks through the inputs
            foreach (ExpressionArgument expression in invocationExpression.Parameters)
            {
                if (expression is PathExpression)
                {
                    sourcePath = ((PathExpression)expression).Path;
                    if (BlockCreator.CompareSourceRootPaths(sourcePath, loopVariable))
                    {
                        return true;
                    }

                    hasPathExpressions = true;
                }
            }

            // returns false if none of the input paths from source are from current scope
            // else it returns true as it does not have input paths from source. Hence it might be a constant value and so must be in the current scope 
            return !hasPathExpressions;
        }

        // checks if the functoid is in current scope
        private static bool FunctoidInCurrentBlock(FunctionInvocationExpression invocationExpression, string loopVariable)
        {
            string sourcePath = string.Empty;
            bool hasPathExpressions = false;

            // checks through the inputs
            foreach (ExpressionArgument expression in invocationExpression.Parameters)
            {
                if (expression is PathExpression)
                {
                    sourcePath = ((PathExpression)expression).Path;
                    if (BlockCreator.CompareSourceRootPaths(sourcePath, loopVariable))
                    {
                        return true;
                    }

                    hasPathExpressions = true;
                }
            }

            // for cumulative functoids, checks through the lists and sources which are populated in the target
            if (invocationExpression.Target != null)
            {
                hasPathExpressions = true;
                sourcePath = ((ListPathExpression)invocationExpression.Target).Path;
                if (BlockCreator.CompareSourceRootPaths(sourcePath, loopVariable))
                {
                    return true;
                }
            }

            // returns false if none of the input paths from source are from current scope
            // else it returns true as it does not have input paths from source. Hence it might be a constant value and so must be in the current scope 
            return !hasPathExpressions;
        }

        // returns true if the source root path is same as that of the current scope loop variable
        private static bool CompareSourceRootPaths(string sourcePath, string loopVariable)
        {
            if (!sourcePath.StartsWith("src"))
            {
                return true; // This is the case where the inputs are not paths from the source tree
            }

            string sourceRoot = sourcePath;
            if (sourcePath.IndexOf('/') >= 0)
            {
                sourceRoot = sourcePath.Substring(0, sourcePath.IndexOf('/'));
            }
            
            return sourceRoot.Equals(loopVariable);
        }
    }
}
