﻿//-----------------------------------------------------------------------
// <copyright file="ForEachHandler.cs" company="Microsoft">
//     Copyright (c) Microsoft. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

namespace BtmToTrfmConvertor
{
    using System.Collections.Generic;
    using System.Linq;
    using System.Xml;
    using Logger;
    using Microsoft.ApplicationServer.Integration.Transforms.Common.Types;
    using Microsoft.ApplicationServer.Integration.Transforms.MapOM;

    public class ForEachHandler : IXslElementHandler
    {
        bool mapeach;
        XmlReader reader = null;
        OperationContext newContext;
        OperationContext parentContext;
        bool emptyLoop;
        Block loop;
        int totalCount;

        // initialises the handler
        public ForEachHandler(XmlReader xmlReader, OperationContext operationContext)
        {
            this.mapeach = true;
            this.reader = xmlReader;
            this.parentContext = operationContext;
            this.newContext = new OperationContext(operationContext);
            this.emptyLoop = true;
        }

        // Handles conversion of the subtree into a trfm mapeach structure
        public object HandleElementConversion()
        {
            Controller.log.trace(MsgType.INFO, "Entering a looping construct");
            string keyMapeach;
            int foreachCounterValue;

            // Source Path Value
            this.reader.Read();
            this.reader.MoveToAttribute(Constants.XslConstants.SelectAttribute);
            string sourceNodeReaderValue;
            if (this.reader.Value.StartsWith(Constants.sourceTemplate))
            {
                sourceNodeReaderValue = this.reader.Value.Substring(Constants.sourceTemplate.Length);
            }
            else
            {
                sourceNodeReaderValue = this.reader.Value;
            }

            ListPathExpression sourcePathValue = this.GetSourcePathValue(sourceNodeReaderValue);
            if (sourcePathValue == null)
            {
                return null;
            }

            this.UpdateKeysAndCounters(out keyMapeach, out foreachCounterValue);
            this.UpdateContext(keyMapeach);
            string targetNode = PathCreation.CreateTargetNode(this.newContext.TargetScope);
            if (targetNode == null)
            {
                Controller.log.debug(MsgType.ERROR, "target node path is null");
                Controller.log.debug(MsgType.UNSUCCESSFUL_CONVERSION, "Looping construct failed to be converted");
                return null;
            }         

            Controller.log.trace(MsgType.INFO, "Parsing the nodes inside the Loop");
            this.Parse();

            if (this.emptyLoop)
            {
                // empty loop is not to be created
                // these are usually created when there are cumulative functoids. Hence appending the whole value to the cumulative input
                if (!Constants.CumulativeInput.StartsWith("$var"))
                {
                    Constants.CumulativeInput = sourceNodeReaderValue + "/" + Constants.CumulativeInput;
                }

                this.parentContext.CurrentListInitialisation = null; // list creation is not necessary
                this.ReassignGlobalVariables(foreachCounterValue);
                return null; 
            }
            else if (this.mapeach)
            {
                this.loop = this.CreateMapStatement(this.loop, keyMapeach, sourcePathValue, this.GetTargetPathValue(keyMapeach, targetNode));
            }
            else
            {
                this.loop = this.CreateMapStatement(this.loop, Constants.ForeachLabel, sourcePathValue);
            }

            return this.loop;      
        }

        // parses the subtree under the for each node
        public void Parse()
        {
            XslReader xslReader = new XslReader(this.reader);
            xslReader.Read(this.newContext); // reads through the xml subtree and updates the currentcontext
            this.mapeach = xslReader.Mapeach;
            Controller.log.trace(MsgType.INFO, "Parsing of nodes under the loop completed");
            if (this.mapeach)
            {
                this.loop = new Mapeach();
            }
            else
            {
                this.loop = new Foreach();
            }

            this.Convert();
        }

        // adds the objects found in the subtree to the mapstatement
        public void Convert()
        {
            List<object> statements = this.newContext.GetItemsInCurrentContext();
            Controller.log.trace(MsgType.INFO, "adding constructs found under the loop");
            if (statements.Count < 2 && statements[0].GetType() == typeof(AddItemToList))
            {
                return; // everytime a list with additemToList is created for a cumulative functoid. It might not be necessary
            }
            else
            {
                this.emptyLoop = false;
            }

            string loopVariable = Constants.TRFMConstants.ScopeSourcePrefix + this.totalCount;
            foreach (object statement in statements)
            {
                if (BlockCreator.BelongsToCurrentBlock(statement, loopVariable))
                {
                    Controller.log.trace(MsgType.INFO, "object of type " + statement.GetType().ToString() + "being added to mapeach");
                    if (statement.GetType() == typeof(PathExpression) || statement.GetType() == typeof(ConstantExpression))
                    {
                        // conditional looping
                        this.loop = this.AddConditionToLoop(this.loop, statement);
                    }
                    else
                    { 
                        BlockCreator.AddToBlock(ref this.loop, statement);
                    }
                }
                else
                {
                    this.parentContext.AddItemToCurrentContext(statement);
                }
            }
        }

        #region private functions

        private void UpdateKeysAndCounters(out string keyMapeach, out int foreachCounterValue)
        {
            keyMapeach = Constants.TRFMConstants.MapeachPrefix + ++TrfmCounters.MapEachCounter;
            this.totalCount = TrfmCounters.MapEachCounter + TrfmCounters.ForEachCounter;
            foreachCounterValue = TrfmCounters.ForEachCounter;
        }

        private void ReassignGlobalVariables(int foreachCounterValue)
        {
            this.newContext.RemoveScopeAndLoopVariablesWithValue(Constants.TRFMConstants.ScopeTagetPrefix + this.totalCount);
            if (foreachCounterValue == TrfmCounters.ForEachCounter)
            {
                TrfmCounters.MapEachCounter--;
            }
            else
            {
                TrfmCounters.ForEachCounter--;
            }
        }

        // Updates the context variables to the current context
        private void UpdateContext(string keyMapeach)
        {
            if (!this.HasAForeachParent())
            {
                this.newContext.TargetScope = this.newContext.SubTreePath;
            }

            this.UpdateContextTargetScope();
            this.newContext.Scope = this.newContext.Scope + "/" + keyMapeach;
            this.newContext.SubTreePath = Constants.TRFMConstants.ScopeTagetPrefix + this.totalCount;
            this.newContext.SetScopeAndLoopVariables(keyMapeach, Constants.TRFMConstants.ScopeSourcePrefix + this.totalCount);
        }

        // updates the target scope in case there are other nodes in between 
        private void UpdateContextTargetScope()
        {
            if (this.newContext.SubTreePath.Split('/').Count() == 1)
            {
                return;
            }

            this.newContext.TargetScope += this.newContext.SubTreePath.Substring(this.newContext.SubTreePath.IndexOf('/'));
        }

        // Returns the targetPathValue for the map each loop
        private ListPathExpression GetTargetPathValue(string key, string tgtNode)
        {
            string prefix = string.Empty;
            int tempValue = TrfmCounters.MapDictionary[key].IndexOf(':');
            if (tempValue != -1)
            {
                prefix = PathCreation.GetPrefix(TrfmCounters.MapDictionary[key].Substring(0, tempValue)) + ":";
            }

            return ExpressionArgumentCreator.CreateListPathExpression(tgtNode + "/" + prefix + PathCreation.RemovePrefixFromNode(TrfmCounters.MapDictionary[key]));
        }

        // checks if the parent of the current scope is a for each loop
        private bool HasAForeachParent()
        {
            string temp = string.Empty;
            bool foreachParent = false;
            if (this.newContext.Scope.LastIndexOf('/') > 0)
            {
                temp = this.newContext.Scope.Substring(this.newContext.Scope.LastIndexOf('/') + 1);
                if (temp.StartsWith(Constants.TRFMConstants.ForeachPrefix))
                {
                    foreachParent = true;
                    Constants.ForeachLabel = temp;
                }
            }

            return foreachParent;
        }

        // creates the map statement by updating the properties and the source details for the for each loop
        private Block CreateMapStatement(Block mapStmt, string currentForEachLabel, ListPathExpression srcPathValue)
        {
            Foreach forEachStatement = (Foreach)mapStmt;

            // add the properties to the for each stmt
            forEachStatement.Properties.Add("XCell", Constants.TRFMConstants.XcellValue);
            forEachStatement.Properties.Add("YCell", Constants.TRFMConstants.YcellValue);
            forEachStatement.Properties.Add("Label", currentForEachLabel);
            forEachStatement.Properties.Add("Id", TrfmCounters.GetNewFunctoidId() );

            // source data updation
            forEachStatement.SourceLoopPath = srcPathValue;
            forEachStatement.SourceLoopVariableName = Constants.TRFMConstants.ScopeSourcePrefix + this.totalCount;
            Controller.log.trace(MsgType.INFO, "ForEach :" + forEachStatement.Properties["Label"] + "\n\tSourceLoopVariable:" + forEachStatement.SourceLoopVariableName + "\n\tSource Path:" + srcPathValue.Path);
            Controller.log.trace(MsgType.INFO, "Exiting Loop");
            return forEachStatement;
        }

        // creates the map statement by updating the properties and the source and target details for the map each loop
        private Block CreateMapStatement(Block mapStmt, string keyMapeach, ListPathExpression srcPathValue, ListPathExpression tgtPathValue)
        {
            Mapeach mapEachStatement = (Mapeach)mapStmt;

            // adding properties
            mapEachStatement.Properties.Add("XCell", Constants.TRFMConstants.XcellValue);
            mapEachStatement.Properties.Add("YCell", Constants.TRFMConstants.YcellValue);
            mapEachStatement.Properties.Add("Label", keyMapeach);
            mapEachStatement.Properties.Add("Id", TrfmCounters.GetNewFunctoidId());

            // updating the paths and variable names
            mapEachStatement.SourceLoopPath = srcPathValue;
            mapEachStatement.SourceLoopVariableName = Constants.TRFMConstants.ScopeSourcePrefix + this.totalCount;
            mapEachStatement.TargetLoopPath = tgtPathValue;
            mapEachStatement.TargetLoopVariableName = Constants.TRFMConstants.ScopeTagetPrefix + this.totalCount;

            Controller.log.trace(MsgType.INFO, "MapEach :" + mapEachStatement.Properties["Label"] + "\n\tSourceLoopVariable:" + mapEachStatement.SourceLoopVariableName + "\n\tTargetLoopVariable:" + mapEachStatement.TargetLoopVariableName + "\n\tSource Path:" + mapEachStatement.SourceLoopPath + "\n\tTarget Path:" + mapEachStatement.TargetLoopPath);
            Controller.log.trace(MsgType.INFO, "Exiting Loop");
            return mapEachStatement;
        }

        // Adds the condition to the loop in case there is a condition upon which the loop is to be created
        private Block AddConditionToLoop(Block mapStmt, object obj)
        {
            string condition;
            VariableInitialization conditionFunctoid = null;
            TextBasedExpression conditionExpression = null;

            if (obj is ConstantExpression)
            {
                condition = ((ConstantExpression) obj).Value.ToString();
            }
            else if (obj is PathExpression)
            {
                string sourceFunctoid = ((PathExpression) obj).Path;

                conditionFunctoid = GetConditionFunctoid(sourceFunctoid);
                conditionExpression = conditionFunctoid.Initializer as TextBasedExpression;
                if(conditionFunctoid != null)
                {
                    condition = conditionExpression.ExpressionText;
                }
                else
                {
                    condition = sourceFunctoid;
                }
            }
            else
            {
                Controller.log.debug(MsgType.UNSUCCESSFUL_CONVERSION, "The condition to the the scope was not converted successfully");
                return mapStmt;
            }

            // remove the condition functoid once it has been mapped
            //if (conditionFunctoid != null)
            //    mapStmt.Statements.Remove(conditionFunctoid);

            if (this.mapeach)
            {
                Mapeach mapeachMapStatment = (Mapeach)mapStmt;
                mapeachMapStatment.Where = condition;
                if (conditionFunctoid != null)
                {
                    for (int i = 0; i < conditionExpression.Aliases.Count; i++)
                    {
                        mapeachMapStatment.Aliases.Add(conditionExpression.Aliases[i]);
                    }
                }
                return mapeachMapStatment;
            }
            else
            {
                Foreach foreachMapStatement = (Foreach)mapStmt;
                foreachMapStatement.Where = condition;
                if (conditionFunctoid != null)
                {
                    for (int i = 0; i < conditionExpression.Aliases.Count; i++)
                    {
                        foreachMapStatement.Aliases.Add(conditionExpression.Aliases[i]);
                    }
                }
                return foreachMapStatement; 
            }
        }

        private VariableInitialization GetConditionFunctoid(string functoidName)
        {
            List<object> statements = this.newContext.GetItemsInCurrentContext();
            string loopVariable = Constants.TRFMConstants.ScopeSourcePrefix + this.totalCount;
            foreach (object statement in statements)
            {
                if (BlockCreator.BelongsToCurrentBlock(statement, loopVariable))
                {
                    if (statement is VariableInitialization)
                    {
                        VariableInitialization variable = (VariableInitialization) statement;
                        if (variable.Identifier == functoidName && variable.Initializer is TextBasedExpression)
                        {
                            return variable;
                        }
                    }
                }
            }
            return null;
        }

        // gets the source path value
        private ListPathExpression GetSourcePathValue(string sourceNodeReaderValue)
        {
            string sourceNode = PathCreation.CreateSourceNode(sourceNodeReaderValue, this.newContext);
            if (sourceNode == null)
            {
                Controller.log.debug(MsgType.ERROR, "source node path is null");
                Controller.log.debug(MsgType.UNSUCCESSFUL_CONVERSION, "Looping construct failed to be converted");
                return null;
            }

            return ExpressionArgumentCreator.CreateListPathExpression(sourceNode);
        }

        #endregion
    }
}
