﻿//-----------------------------------------------------------------------
// <copyright file="FunctoidDictionary.cs" company="Microsoft">
//     Copyright (c) Microsoft. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

namespace BtmToTrfmConvertor
{
    using System.Collections.Generic;
    using Microsoft.ApplicationServer.Integration.Transforms.Common;
    using Microsoft.ApplicationServer.Integration.Transforms.MapOM;

    class FunctoidDictionary
    {
        Dictionary<string, string> supportedFunctoids = new Dictionary<string, string>();
        Dictionary<string, string> unsupportedFunctoids = new Dictionary<string, string>();

        public FunctoidDictionary()
        {
            this.BuildSupportedFunctoidsDictionary();
            this.BuildUnsupportedFunctoidsDictionary();
        }

        internal string GetFunctoidName(string value)
        {
            if (this.supportedFunctoids.ContainsKey(value))
            {
                return this.supportedFunctoids[value];
            }

            if (this.unsupportedFunctoids.ContainsKey(value))
            {
                return this.unsupportedFunctoids[value];
            }

            return "Functoid not in dictionary";
        }

        internal bool IsSupported(string value)
        {
            if (this.supportedFunctoids.ContainsKey(value))
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        internal bool ContainsKey(string value)
        {
            if (this.unsupportedFunctoids.ContainsKey(value) || this.supportedFunctoids.ContainsKey(value))
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        #region accecessing function names from trfm source code

        private string GetFunctionName(int functoidID)
        {
            return BuiltinFunctionsHelper.FunctionIdToFunctionName(functoidID).ToString();
        }

        private string GetMathFunctionName(string operation)
        {
            return (char)Types.FunctoidType.ArithmeticExpression + operation;
        }

        private string GetLogicalFunctionName(string operation)
        {
            return (char)Types.FunctoidType.LogicalExpression + operation;
        }

        private string GetDateTimeFunctionName(Types.DateTimeFunctoidType dateTimeFunctoidType, int functoidID)
        {
            return (char)Types.FunctoidType.DateTimeFunctoid + ((char)dateTimeFunctoidType).ToString() + this.GetFunctionName(functoidID);
        }

        private string GetCumulativeFunctionName(int functoidID)
        {
            return (char)Types.FunctoidType.CumulativeFunctoid + this.GetFunctionName(functoidID);
        }

        private string GetStringFunctionName(int functoidID)
        {
            return (char)Types.FunctoidType.StringFunctoid + this.GetFunctionName(functoidID);
        }

        private string GetScriptingFunctionName()
        {
            return (char)Types.FunctoidType.ScriptingFunctoid + "CSharp";
        }

        #endregion

        void BuildSupportedFunctoidsDictionary()
        {
            // 1--String Functoids 
            this.supportedFunctoids.Add("StringUpperCase", this.GetStringFunctionName((int)BaseFunctoidIDs.StringUpperCase));
            this.supportedFunctoids.Add("StringLowerCase", this.GetStringFunctionName((int)BaseFunctoidIDs.StringLowerCase));
            this.supportedFunctoids.Add("StringTrimRight", this.GetStringFunctionName((int)BaseFunctoidIDs.StringTrimRight));
            this.supportedFunctoids.Add("StringTrimLeft", this.GetStringFunctionName((int)BaseFunctoidIDs.StringTrimLeft));
            this.supportedFunctoids.Add("StringSubstring", this.GetStringFunctionName((int)BaseFunctoidIDs.StringSubstring));
            this.supportedFunctoids.Add("StringSize", this.GetStringFunctionName((int)BaseFunctoidIDs.StringSize));
            this.supportedFunctoids.Add("StringRight", this.GetStringFunctionName((int)BaseFunctoidIDs.StringRight));
            this.supportedFunctoids.Add("StringLeft", this.GetStringFunctionName((int)BaseFunctoidIDs.StringLeft));
            this.supportedFunctoids.Add("StringFind", this.GetStringFunctionName((int)BaseFunctoidIDs.StringFind));
            this.supportedFunctoids.Add("StringConcat", this.GetStringFunctionName((int)BaseFunctoidIDs.StringConcatenate));

            // 2--Math Functoids
            this.supportedFunctoids.Add("MathAdd", this.GetMathFunctionName("+"));
            this.supportedFunctoids.Add("MathAbs", this.GetMathFunctionName("Abs"));
            this.supportedFunctoids.Add("MathMod", this.GetMathFunctionName("%"));
            this.supportedFunctoids.Add("MathRound", this.GetMathFunctionName("Round"));
            this.supportedFunctoids.Add("MathSqrt", this.GetMathFunctionName("Sqrt"));
            this.supportedFunctoids.Add("MathSubtract", this.GetMathFunctionName("-"));
            this.supportedFunctoids.Add("MathMultiply", this.GetMathFunctionName("*"));
            this.supportedFunctoids.Add("MathDivide", this.GetMathFunctionName("\\"));

            // 3--Logical Functoids
            this.supportedFunctoids.Add("LogicalGt", this.GetLogicalFunctionName(">"));
            this.supportedFunctoids.Add("LogicalGte", this.GetLogicalFunctionName(">="));
            this.supportedFunctoids.Add("LogicalLt", this.GetLogicalFunctionName("<"));
            this.supportedFunctoids.Add("LogicalLte", this.GetLogicalFunctionName("<="));
            this.supportedFunctoids.Add("LogicalEq", this.GetLogicalFunctionName("=="));
            this.supportedFunctoids.Add("LogicalNe", this.GetLogicalFunctionName("!="));
            this.supportedFunctoids.Add("LogicalOr", this.GetLogicalFunctionName("||"));
            this.supportedFunctoids.Add("LogicalAnd", this.GetLogicalFunctionName("&&"));
            this.supportedFunctoids.Add("LogicalExistence", this.GetLogicalFunctionName("Exists"));
            this.supportedFunctoids.Add("LogicalIsDate", this.GetLogicalFunctionName("IsDate"));
            this.supportedFunctoids.Add("LogicalIsNumeric", this.GetLogicalFunctionName("IsNumber"));
            this.supportedFunctoids.Add("LogicalNot", this.GetLogicalFunctionName("!"));

            // 4--DateTime functoids
            this.supportedFunctoids.Add("DateCurrentDate", this.GetDateTimeFunctionName(Types.DateTimeFunctoidType.DateCurrentDate, (int)BaseFunctoidIDs.DateTimeGenerate));
            this.supportedFunctoids.Add("DateCurrentTime", this.GetDateTimeFunctionName(Types.DateTimeFunctoidType.DateCurrentTime, (int)BaseFunctoidIDs.DateTimeGenerate));
            this.supportedFunctoids.Add("DateCurrentDateTime", this.GetDateTimeFunctionName(Types.DateTimeFunctoidType.DateCurrentDateTime, (int)BaseFunctoidIDs.DateTimeGenerate));

            // 5--Cumulative functoids
            this.supportedFunctoids.Add("GetCumulativeConcat", this.GetCumulativeFunctionName((int)BaseFunctoidIDs.CumulativeConcat));
            this.supportedFunctoids.Add("GetCumulativeAvg", this.GetCumulativeFunctionName((int)BaseFunctoidIDs.CumulativeAverage));
            this.supportedFunctoids.Add("GetCumulativeMax", this.GetCumulativeFunctionName((int)BaseFunctoidIDs.CumulativeMax));
            this.supportedFunctoids.Add("GetCumulativeMin", this.GetCumulativeFunctionName((int)BaseFunctoidIDs.CumulativeMin));
            this.supportedFunctoids.Add("GetCumulativeSum", this.GetCumulativeFunctionName((int)BaseFunctoidIDs.CumulativeSum));
            this.supportedFunctoids.Add("count", this.GetCumulativeFunctionName((int)BaseFunctoidIDs.CumulativeCount));

            // 6--Scripting
            this.supportedFunctoids.Add("CSharp", this.GetScriptingFunctionName());

            // 7--Misc functoids
            this.supportedFunctoids.Add("Iteration", ContextFunctionNames.GetCurrentPosition.ToString());
        }

        void BuildUnsupportedFunctoidsDictionary()
        {
            // scrapped functoids
            this.unsupportedFunctoids.Add("MathInt", "Integer Functoid");
            this.unsupportedFunctoids.Add("MathMax", "Maximum Value Functoid");
            this.unsupportedFunctoids.Add("MathMin", "Minimum Value Functoid");
            this.unsupportedFunctoids.Add("LogicalIsString", "IsString Functoid");
            this.unsupportedFunctoids.Add("DateAddDays", "Add Days Functoid");

            // yet to be implemented functoids
            //
            // scientific functoids
            this.unsupportedFunctoids.Add("SciArcTan", "Arc Tangent Functoid");
            this.unsupportedFunctoids.Add("SciCos", "Cosine Functoid");
            this.unsupportedFunctoids.Add("SciPow", "X ^ Y Functoid");
            this.unsupportedFunctoids.Add("SciSin", "Sine Functoid");
            this.unsupportedFunctoids.Add("SciTan", "Tangent Functoid");
            this.unsupportedFunctoids.Add("SciExp", "Natural Exponential Functoid");
            this.unsupportedFunctoids.Add("SciExp10", "10 ^ n functoid");
            this.unsupportedFunctoids.Add("SciLog10", "Common logarithm functoid");
            this.unsupportedFunctoids.Add("SciLog", "Natural logarithm functoid");
            this.unsupportedFunctoids.Add("SciLogn", "Base specified Logarithm functoid");

            // conversion functoids
            this.unsupportedFunctoids.Add("ConvertAsc", "Character To ASCII Functoid");
            this.unsupportedFunctoids.Add("ConvertChr", "ASCII To Character Functoid");
            this.unsupportedFunctoids.Add("ConvertHex", "Hexadecimal Functoid");
            this.unsupportedFunctoids.Add("ConvertOct", "Octal Functoid");

            // Database Functoid
            this.unsupportedFunctoids.Add("DBLookup", "Database Lookup Functoid");
            this.unsupportedFunctoids.Add("DBErrorExtract", "Error Return Functoid");
            this.unsupportedFunctoids.Add("FormatMessage", "Format Message Functoid");
            this.unsupportedFunctoids.Add("GetAppID", "Get Application ID Functoid");
            this.unsupportedFunctoids.Add("GetAppValue", "Get Application Value Functoid");
            this.unsupportedFunctoids.Add("GetCommonID", "Get Common ID Functoid");
            this.unsupportedFunctoids.Add("RemoveAppID", "Remove Application ID Functoid");
            this.unsupportedFunctoids.Add("SetCommonID", "Set Common ID Functoid");
            this.unsupportedFunctoids.Add("DBValueExtract", "Value Extractor Functoid");
            this.unsupportedFunctoids.Add("GetCommonValue", "Get Common Value Functoid");
        }
    }
}
