﻿//-----------------------------------------------------------------------
// <copyright file="XSLTGenerator.cs" company="Microsoft">
//     Copyright (c) Microsoft. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

namespace BtmToTrfmConvertor.MapConvertor
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Linq;
    using System.Text;
    using Logger;
    using Microsoft.BizTalk.BaseFunctoids;
    using Microsoft.BizTalk.Mapper.Compiler;
    using Microsoft.BizTalk.Mapper.OM;
    using Microsoft.BizTalk.TOM;

    public class XsltGenerator
    {
        public static string GenerateCustomisedXSLT(string inputBtmFile)
        {
            string newBtmPath = RemoveInlineXSLTFunction(inputBtmFile);
            string xsltPath = CompileBTM(newBtmPath);
            if (File.Exists(newBtmPath))
            {
                File.Delete(newBtmPath);
            }
            return xsltPath;   
        }

        // This function compiles a BTM file into xsl and returns xslfile path
        internal static string CompileBTM(string btmFileWithFullPath)
        {
            XsltCompiler compiler = GetNewXsltCompiler();
            
            compiler.CompileXslt(null, null, btmFileWithFullPath);

            string xslt = compiler.Xslt;

            // create an xslt file in the same location as the btm input file
            string xslFileName = Path.Combine(Path.GetDirectoryName(btmFileWithFullPath), Guid.NewGuid().ToString() + ".xsl");
            Controller.log.debug(MsgType.INFO, "XSLT:" + xslt);
            try
            {
                using (StreamWriter writer = new StreamWriter(xslFileName, false, Encoding.Unicode))
                {
                    writer.Write(xslt);
                }
            }
            catch (Exception e)
            {
                Controller.log.debug(MsgType.UNSUCCESSFUL_CONVERSION, "Unable to compile BTM\nERROR:" + e.Message);
            }

            return xslFileName;
        }

        /// <summary>
        /// Create a new XSLT compiler and intialises it
        /// </summary>
        /// <returns>the xslt compiler created</returns>
        private static XsltCompiler GetNewXsltCompiler()
        {
            XsltCompiler compiler = new XsltCompiler();

            compiler.EnableUnitTesting = true;
            compiler.CompilerErrorHandler = new CompilerErrorManager(2);
            FunctoidManager functoidManager = new FunctoidManager();
            functoidManager.LoadWrappers(null);
            compiler.SetFunctoidManager(functoidManager);

            compiler.SchemaResolver = new SchemaResolver();

            return compiler;
        }

        /// <summary>
        /// This function is going to return the path of the new btm file that will be saved. The name of the new bt, file
        /// will be : oldbtm Name + _bak.btm
        /// Please note that the input parameter that is being passed should have Schemas that can be loaded as defined in the btm file
        /// For example if the btm has : SourceSchema : .\ABC.xsd then ABC.xsd should be present in the same folder
        /// </summary>
        /// <returns></returns>
        private static string RemoveInlineXSLTFunction(string btmFilePath)
        {
            const string InlineXsltFunctionCall = "inlineXsltFunctionCall";
            const string InlineXsltCallTemplateFunctionCall = "InlineXsltCallTemplateFunctionCall";

            Map map = new Map();
            map.LoadFunctoids(null);
            map.Load(btmFilePath);

            int counter1 = 1;
            int counter2 = 1;

            // retrievinng the schema paths from the BTM File
            string TargetSchemaPath = Path.GetFullPath(map.TargetTreePath);
            string SourceSchemaPath = Path.GetFullPath(map.SourceTreePath);
            foreach (var page in map.Pages)
            {
                // removing scripting functoids
                foreach (var scriptingFunctoid in page.Functoids.Where((f) => f.FunctoidID == 260))
                {
                    string xsltData = scriptingFunctoid.BaseFunctoid.GetScriptBuffer(ScriptType.Xslt);
                    int paramsCount = scriptingFunctoid.InputLinks.Count;

                    // Generates CSV separated params
                    string paramText = GenerateScriptFunctoidParameters(paramsCount);
                    if (!string.IsNullOrEmpty(xsltData))
                    {
                        string newData = "public void " + InlineXsltFunctionCall + counter1 + "(" + paramText + "){ /*" + xsltData + "*/}";
                        scriptingFunctoid.BaseFunctoid.SetScriptBuffer(ScriptType.CSharp, newData);
                        counter1++;
                    }

                    xsltData = scriptingFunctoid.BaseFunctoid.GetScriptBuffer(ScriptType.XsltCallTemplate);
                    if (!string.IsNullOrEmpty(xsltData))
                    {
                        string newData = "public void " + InlineXsltCallTemplateFunctionCall + counter2 + "(" + paramText + "){ /*" + xsltData + "*/}";
                        scriptingFunctoid.BaseFunctoid.SetScriptBuffer(ScriptType.CSharp, newData);
                        counter2++;
                    }
                }
            }

            string newFileName = Path.Combine(Path.GetDirectoryName(btmFilePath), Path.GetFileNameWithoutExtension(btmFilePath) + "_temp.btm");
            map.Save(newFileName);
            return newFileName;
        }

        private static string GenerateScriptFunctoidParameters(int paramsCount)
        {
            string paramText = string.Empty;
            int i = 0;
            while (i++ < paramsCount)
            {
                paramText += "string dummy" + i.ToString();
                if (i != paramsCount)
                {
                    paramText += ",";
                }
            }

            return paramText;
        }
    }
}
