﻿//-----------------------------------------------------------------------
// <copyright file="NodeVisitor.cs" company="Microsoft">
//     Copyright (c) Microsoft. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

namespace BtmToTrfmConvertor
{
    using System;
    using System.Collections.Generic;
    using Logger;
    using Microsoft.ApplicationServer.Integration.Transforms.Common.Types;
    using Microsoft.ApplicationServer.Integration.Transforms.MapOM;
    using Microsoft.ApplicationServer.Integration.Transforms.MapOM.Parsing;

    public class NodeVisitor : INodeVisitor
    {
        XslExpressionHandler expressionHandler;

        // based on the caller creates a LogicalExpression,ArithmeticExpression or a FunctoidIncvocation
        public object Visit(INodeVisitable v)
        {
            OperationContext context;
            int charCut;
            string inputs, functionDetail, functionTarget,content;
            Type visitorType = v.GetType();
            content = null;
            if (visitorType == typeof(LogicalExpressionCreator))
            {
                LogicalExpressionCreator creator = (LogicalExpressionCreator)v;
                functionDetail = creator.FunctionDetail;
                functionTarget = creator.FunctionTarget;
                context = creator.Context;
                Controller.log.trace(MsgType.INFO, "Logical Expression detected");
            }
            else if (visitorType == typeof(ArithmeticExpressionCreator))
            {
                ArithmeticExpressionCreator creator = (ArithmeticExpressionCreator)v;
                functionDetail = creator.FunctionDetail;
                functionTarget = creator.FunctionTarget;
                context = creator.Context;
                Controller.log.trace(MsgType.INFO, "Arithmetic Expression detected");
            }
            else if (visitorType == typeof(FunctoidInvocationCreator))
            {
                FunctoidInvocationCreator creator = (FunctoidInvocationCreator)v;
                functionDetail = creator.FunctionDetail;
                functionTarget = creator.FunctionTarget;
                context = creator.Context;
                Controller.log.trace(MsgType.INFO, "Functoid Invocation detected");
            }
            else
            {
                ScriptInvocationCreator creator = (ScriptInvocationCreator)v;
                functionDetail = creator.FunctionDetail;
                functionTarget = creator.FunctionTarget;
                context = creator.Context;
                content = creator.Content;
                Controller.log.trace(MsgType.INFO, "Script Invocation detected");
            }

            // removing the curly braces around the inputs ---- (inputs) -> inputs
            charCut = functionDetail.IndexOf('(');
            inputs = functionDetail.Substring(charCut + 1, functionDetail.Length - (2 + charCut));
            this.expressionHandler = new XslExpressionHandler(context);
            if (this.IsOperator(visitorType, functionTarget))
            {
                Controller.log.trace(MsgType.INFO, "An expression is created by using operators between inputs");
                string expression = this.GenerateExpression(inputs, functionTarget);
                return this.expressionHandler.ParseExpression(expression);
            }
            else
            {
                Controller.log.trace(MsgType.INFO, "An expression is created by using function calls");

                // PROBLEM IF THE CONSTANT HAS " , "
                List<object> inputList = this.ProcessInputs(inputs);
                if (visitorType == typeof(FunctoidInvocationCreator))
                {
                    return this.ProcessFunctionInvocation(inputList, functionTarget, context);
                }
                else if (visitorType == typeof(ScriptInvocationCreator))
                {
                    return this.ProcessScriptInvocation(inputList, functionTarget, context, content);
                }
                else
                {
                    List<NameBinding> nameBindings = new List<NameBinding>();
                    int index = 0;
                    foreach (object input in inputList)
                    {
                        index++;
                        nameBindings.Add(this.expressionHandler.formatInput(input, "Input_" + index, getTypeBasedOnFunctionName(functionTarget)));
                    }

                    if (nameBindings == null)
                    {
                        return null;
                    }

                    if (visitorType == typeof(LogicalExpressionCreator))
                    {
                        return this.ProcessLogicalExpressionWithFunctions(nameBindings, functionTarget);
                    }
                    else
                    {
                        return this.ProcessArithmeticExpressionWithFunctions(nameBindings, functionTarget);
                    }
                }                
            }
        }

        private object ProcessScriptInvocation(List<object> inputList, string functionTarget, OperationContext context, string content)
        {
            ScriptInvocationExpression scriptInvocationExpression = new ScriptInvocationExpression();
            foreach (object input in inputList)
            {
                scriptInvocationExpression.Parameters.Add(input as ExpressionArgument);
            }

            scriptInvocationExpression.ScriptId = TrfmCounters.GetNewFunctoidId();
            Constants.map1.Scripts.Add(new Script() { ScriptId = scriptInvocationExpression.ScriptId, Language = "C#", Content = content });
            return scriptInvocationExpression;
        }

        private ITransformType getTypeBasedOnFunctionName(string functionName)
        {
            MapFunctionType functionType;
            List<MapFunctionType> mathFunctions = new List<MapFunctionType>();
            mathFunctions.Add(MapFunctionType.Abs);
            mathFunctions.Add(MapFunctionType.Round);
            mathFunctions.Add(MapFunctionType.Min);
            mathFunctions.Add(MapFunctionType.Max);
            mathFunctions.Add(MapFunctionType.Sqrt);
            if (MapFunctions.TryGetFunction(functionName, out functionType))
            {
                if (mathFunctions.Contains(functionType))
                {
                    return SimpleType.Number;
                }
            }
            return SimpleType.String;
        }

        // Processes the inputs to function invocation
        private object ProcessFunctionInvocation(List<object> input, string funcTgt, OperationContext context)
        {
            FunctionInvocationExpression funcInvoke = new FunctionInvocationExpression();
            foreach (object inputValue in input)
            {
                ExpressionArgument processedInput = (ExpressionArgument)this.ProcessFunctionInvocationInput(inputValue, funcTgt);
                if (funcTgt.Contains("Cumulative"))
                {
                    ITransformType returnType;
                    funcInvoke.Target = processedInput;
                    if (funcTgt.Contains("Concat"))
                    {
                        returnType = SimpleType.String;
                    }
                    else
                    {
                        returnType = SimpleType.Number;
                    }

                    if (context.CurrentListInitialisation != null)
                    {
                        funcInvoke.Parameters.Add(ExpressionArgumentCreator.CreateLambdaExpression("item.Input_0", new List<string> { "item" }, returnType));
                    }
                }
                else
                {
                    funcInvoke.Parameters.Add(processedInput);
                }
            }

            funcInvoke.FunctionName = funcTgt;
            Controller.log.trace(MsgType.INFO, "The function name is :" + funcInvoke.FunctionName);
            return funcInvoke;
        }

        // Processes the inputs to logical expression
        private object ProcessLogicalExpressionWithFunctions(List<NameBinding> nameBindings, string funcTgt)
        {
            LogicalExpression logicalExp = new LogicalExpression();
            return this.GenerateExpressionFunctoid(logicalExp, nameBindings, funcTgt);
        }

        // Processes the inputs to arithmetic expression after handling the exceptions such as Round functoid
        private object ProcessArithmeticExpressionWithFunctions(List<NameBinding> nameBindings, string funcTgt)
        {
            ArithmeticExpression arithmeticExp = new ArithmeticExpression();
            string aliases = string.Empty;
            switch (funcTgt)
            {
                case "Round":
                    ConstantExpression constExp = (ConstantExpression)nameBindings[1].Expression;
                    aliases = "Input_1," + constExp.Value;
                    arithmeticExp.Aliases.Add(nameBindings[0]);
                    arithmeticExp.ExpressionText = funcTgt + "(" + aliases + ")";
                    break;

                case "Notes":
                    int inputCount = nameBindings.Count;
                    for (int j = 0; j < inputCount; j++)
                    {
                        aliases = aliases + "Input_" + (j + 1) + ",";
                        arithmeticExp.Aliases.Add(nameBindings[j]);
                    }

                    arithmeticExp.ExpressionText = string.Empty;
                    break;

                default:
                    return this.GenerateExpressionFunctoid(arithmeticExp, nameBindings, funcTgt);
            }

            Controller.log.trace(MsgType.INFO, "The function name is :" + funcTgt);
            Controller.log.trace(MsgType.INFO, "The expression is " + arithmeticExp.ExpressionText);
            return arithmeticExp;
        }

        // Generates an Expression functoid by adding the inputs along with the aliases and creating the expression text
        private TextBasedExpression GenerateExpressionFunctoid(TextBasedExpression expression, List<NameBinding> nameBindings, string funcTgt)
        {
            string aliases = string.Empty;
            TextBasedExpression expressionFunctoid = expression;
            int inputCount = nameBindings.Count;
            for (int j = 0; j < inputCount; j++)
            {
                aliases = aliases + "Input_" + (j + 1) + ",";
                expressionFunctoid.Aliases.Add(nameBindings[j]);
            }

            expressionFunctoid.ExpressionText = funcTgt + "(" + aliases.Substring(0, aliases.Length - 1) + ")";
            Controller.log.trace(MsgType.INFO, "The function name is :" + funcTgt);
            Controller.log.trace(MsgType.INFO, "The expression is " + expressionFunctoid.ExpressionText);
            return expressionFunctoid;
        }

        // Processes the inputs to a FunctionInvocationExpression and returns an Expression Argument
        private object ProcessFunctionInvocationInput(object input, string funcTgt)
        {
            if (input == null)
            {
                input = this.expressionHandler.CreateNullPath();
            }

            if (funcTgt.Equals("DateTimeGenerate"))
            {
                Controller.log.trace(MsgType.INFO, "DataTime Generate function is encountered");
                ConstantExpression tempExp = (ConstantExpression)input;
                tempExp.Properties.Add("SettingName", "OutputFormat");
                return tempExp;
            }
            else
            {
                return input;
            }
        }

        private List<object> ProcessInputs(string inputs)
        {
            string[] array = { " , " };
            string[] input = inputs.Split(array, StringSplitOptions.RemoveEmptyEntries);
            List<object> inputObj = new List<object>();
            foreach (string temp in input)
            {
                inputObj.Add(this.expressionHandler.ParseExpression(temp));
            }

            return inputObj;
        }

        // Given the inputs and the operator, this function replaces the commas by the operator and returns the expression formed
        private string GenerateExpression(string inputs, string expOperator)
        {
            string expression = string.Empty;
            if (expOperator.Equals("!"))
            {
                expression = "!" + inputs;
            }
            else
            {
                expression = inputs.Replace(" , ", expOperator);
            }

            return expression;
        }

        private bool IsOperator(Type visitorType, string funcTgt)
        {
            if (visitorType == typeof(LogicalExpressionCreator) && funcTgt.Length <= 2)
            {
                return true;
            }

            if (visitorType == typeof(ArithmeticExpressionCreator) && funcTgt.Length == 1)
            {
                return true;
            }

            return false;
        }
    }
}
