﻿//-----------------------------------------------------------------------
// <copyright file="PathCreation.cs" company="Microsoft">
//     Copyright (c) Microsoft. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

namespace BtmToTrfmConvertor
{
    using System;
    using Logger;
    using Microsoft.ApplicationServer.Integration.Transforms.MapOM;

    class PathCreation
    {
        private enum XpathType
        {
            Attribute = '@',
            ConstantValue = '"',
            Variable = '$',
            GlobalScope = '/',
            ParentScope = '.',
            IndexFunctiod = '['
        }

        // creates the source node path in the trfm fle
        internal static string CreateSourceNode(string path, OperationContext context)
        {
            string sourceNode, processedNode, nodeWithoutIndex;
            if (PathCreation.HasIndex(path, out nodeWithoutIndex))
            {
                path = "[" + nodeWithoutIndex;
            }

            XpathType pathType = (XpathType)path.Trim()[0];
            switch (pathType)
            {
                case XpathType.IndexFunctiod:
                    // if the xpath has an index value, the xpath is mapped to an index functoid(arithmetic functoid as it is not yet implemented) in trfm
                    return PathCreation.AddIndexFunctoid(path.Substring(1), context);

                case XpathType.ConstantValue:
                    // removing the quotes and returning the constant value
                    string constantValue = PathCreation.GetConstantValue(path);
                    if (constantValue != null)
                    {
                        return constantValue;
                    }

                    break;

                case XpathType.Variable:
                    return PathCreation.GetTrfmPathForVariable(path, context);
            }

            sourceNode = PathCreation.GetSourceLoopVariableOfScope(path, context);
            
            // Getting the separate nodes from the xpath
            char[] splitChar = { '/' };
            string[] nodes = path.Split(splitChar, StringSplitOptions.RemoveEmptyEntries);
            foreach (string currentNode in nodes)
            {
                processedNode = PathCreation.ProcessNode(currentNode);
                if (processedNode == null) 
                {
                    // in case of nillable nodes
                    return "nil:" + sourceNode;
                }

                sourceNode += processedNode;
            }

            Controller.log.trace(MsgType.INFO, "Source Path generated :" + sourceNode);
            return sourceNode;
        }

        // Gets the TRFM prefix that has to be used based on the prefix used in BTM
        internal static string GetPrefix(string givenPrefix)
        {
            if (!Constants.PrefixDictionary.ContainsKey(givenPrefix))
            {
                Constants.PrefixDictionary.Add(givenPrefix, "ns" + Constants.PrefixDictionary.Count);
            }

            return Constants.PrefixDictionary[givenPrefix];
        }

        // creates the target node path in the trfm file
        internal static string CreateTargetNode(string path)
        {
            string targetNode = string.Empty;
            string start = string.Empty;
            bool firstNode = true;

            // separate the nodes in the incoming path
            string[] paths = path.Split('/');
            foreach (string currentNode in paths)
            {
                if (firstNode)
                {
                    // the first node is taken as it is
                    start = currentNode;
                    firstNode = false;
                }
                else
                {
                    targetNode += PathCreation.CalcPrefix(currentNode, false) + PathCreation.RemovePrefixFromNode(currentNode);
                }
            }

            targetNode = start + targetNode;
            Controller.log.trace(MsgType.INFO, "Target Path generated :" + targetNode);
            return targetNode;
        }

        // remove the prefix from a single node name(example ns0:parent gives parent)
        internal static string RemovePrefixFromNode(string node)
        {
            int charCut = node.IndexOf(':');
            return node.Substring(charCut + 1);
        }

        #region private functions

        // Based on the xpath, the function determines the scope and gives the SourceLoopVariable from TRFM 
        private static string GetSourceLoopVariableOfScope(string path, OperationContext context)
        {
            string sourceLoopVariable = string.Empty;
            string start;
            XpathType pathType = (XpathType)path[0];
            int level;
            switch (pathType)
            {
                case XpathType.GlobalScope:
                    // if the path is from the root node
                    // removing the root node from the path
                    path = path.Trim().Substring(1);
                    path = path.Substring(path.IndexOf("/") + 1);
                    return Constants.TRFMConstants.ScopeSourceHome;

                case XpathType.ParentScope:
                    // if the path denotes to a parent scope or node
                    string currentScope;
                    level = PathCreation.GetHierarchyLevel(path);
                    currentScope = PathCreation.GetScopeAtLevel(context.Scope, level);
                    switch (currentScope)
                    {
                        case null:
                            if (context.SubTreePath.IndexOf('/') >= 0)
                            {
                                start = context.SubTreePath.Substring(0, context.SubTreePath.IndexOf('/'));
                            }
                            else
                            {
                                start = context.SubTreePath;
                            }

                            sourceLoopVariable = "src" + start.Substring(3, start.Length - 3);
                            break;
                        case Constants.TRFMConstants.ScopeHome:
                            sourceLoopVariable = Constants.TRFMConstants.ScopeSourceHome;
                            break;
                        default:
                            sourceLoopVariable = context.GetScopeAndLoopVariables(currentScope);
                            break;
                    }
                    
                    break;

                default:
                    if (context.SubTreePath.IndexOf('/') >= 0)
                    {
                        start = context.SubTreePath.Substring(0, context.SubTreePath.IndexOf('/'));
                    }
                    else
                    {
                        start = context.SubTreePath;
                    }

                    sourceLoopVariable = "src" + start.Substring(3, start.Length - 3);
                    break;
            }

            return sourceLoopVariable;
        }

        private static string GetScopeAtLevel(string scopeHierarchy, int level)
        {
            string[] scope = scopeHierarchy.Split('/');
            Array.Reverse(scope);
            foreach (string currentScope in scope)
            {
                if (level == 0)
                {
                    return currentScope;
                }

                level--;
            }

            return null;
        }

        // Calculates to which parent the xpath points to
        // for example ../../ will point to the grand parent
        // This function returns the number of levels the xpath should go up to reach the current scope
        private static int GetHierarchyLevel(string path)
        {
            int level = 0;
            int count = 0;
            do
            {
                while (path[count] == '.')
                {
                    count++;
                }

                level += count - 1;
                if (count + 1 < path.Length)
                {
                    path = path.Substring(count + 1).Trim();
                }

                count = 0;
            }
            while (path.StartsWith("."));
            return level;
        }

        private static string GetConstantValue(string path)
        {
            if ((path.LastIndexOf('"') - path.IndexOf('"')) > 0)
            {
                return path.Substring(path.IndexOf('"') + 1, path.LastIndexOf('"') - path.IndexOf('"') - 1);
            }

            return null;
        }

        // returns the TRFM path that corresponds to the given variable
        private static string GetTrfmPathForVariable(string path, OperationContext context)
        {
            string pathWithoutPrefix = PathCreation.RemovePrefixFromNode(path);
            if (context.ContainsKey(pathWithoutPrefix))
            {
                return context.GetVarDictionary(pathWithoutPrefix);
            }

            Controller.log.debug(MsgType.ERROR, "Path expression has a variable which has not been logged to the dictionary");
            return null;
        }

        // processes the node from the Xpath
        private static string ProcessNode(string currentNode)
        {
            string prefix = string.Empty;
            string processedValue = string.Empty;
            XpathType pathType = (XpathType)currentNode[0];
            if (!currentNode.Contains("text()") && pathType != XpathType.ParentScope)
            {
                prefix = PathCreation.CalcPrefix(currentNode, true);
                if (prefix == null)
                {
                    return null;
                }

                processedValue = prefix + PathCreation.RemovePrefixFromNode(currentNode);
            }

            return processedValue;
        }

        // calculates the prefix for each node
        private static string CalcPrefix(string currentNode, bool isSourcePath)
        {
            string prefix = "/"; // default value without a preifx
            XpathType pathType = (XpathType)currentNode[0];
            if (isSourcePath && pathType == XpathType.Attribute)
            {
                if (currentNode.Equals("@xsi:nil"))
                {
                    return null;
                }
            }
            else if (currentNode.IndexOf(':') != -1)
            {
                prefix += PathCreation.GetPrefix(currentNode.Substring(0, currentNode.IndexOf(':'))) + ":"; // if the BTM file has a prefix, a corresponding prefix is added in the TRFM file
            }

            return prefix;
        }

        // Checks whether the xpath denotes to some particular index of a node(example-x/y[1]/z)
        // returns the node value without the index part in the path
        private static bool HasIndex(string value, out string nodeWithoutIndex)
        {
            int startIndex, endIndex, intValue;
            nodeWithoutIndex = value;
            startIndex = value.LastIndexOf('[');
            if (startIndex == -1)
            {
                return false;
            }

            endIndex = value.LastIndexOf(']');
            if (endIndex < startIndex + 1)
            {
                return false;
            }

            nodeWithoutIndex = value.Substring(0, startIndex);
            if (endIndex + 1 < value.Length)
            {
                nodeWithoutIndex += value.Substring(endIndex + 1, value.Length - endIndex - 1);
            }

            string temp = value.Substring(startIndex + 1, endIndex - startIndex - 1);
            while (PathCreation.HasIndex(nodeWithoutIndex, out nodeWithoutIndex)) 
            { 
            }

            return int.TryParse(temp, out intValue);
        }

        // adds index fucntoid to the current context
        private static string AddIndexFunctoid(string path, OperationContext context)
        {
            VariableHandler handler = new VariableHandler(context);
            VariableInitialization varInitNode = handler.AddFunctoid("Index functoid", "(" + path + ")", true);
            context.AddItemToCurrentContext(varInitNode);
            return varInitNode.Identifier;
        }

        #endregion
    }
}
