﻿//-----------------------------------------------------------------------
// <copyright file="VariableHandler.cs" company="Microsoft">
//     Copyright (c) Microsoft. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

namespace BtmToTrfmConvertor
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Xml;
    using BtmToTrfmConvertor.MapConvertor;
    using CDataReaderForXSL;
    using Logger;
    using Microsoft.ApplicationServer.Integration.Transforms.Common.Types;
    using Microsoft.ApplicationServer.Integration.Transforms.MapOM;

    public class VariableHandler : IXslElementHandler
    {
        XmlReader reader = null;
        OperationContext context;
        XslExpressionHandler expressionHandler;

        public VariableHandler(XmlReader xmlReader, OperationContext operationContext)
        {
            this.reader = xmlReader;
            this.context = operationContext;
            this.expressionHandler = new XslExpressionHandler(this.context);
        }

        public VariableHandler(OperationContext operationContext)
        {
            this.context = operationContext;
            this.expressionHandler = new XslExpressionHandler(this.context);
        }

        // creates a functoid or stores the variable value in the varDictionary
        public object HandleElementConversion()
        {
            string variableName, variableValue, value, cdataSection;
            VariableInitialization variableNode = null;
            this.reader.Read();
            this.reader.MoveToAttribute(Constants.XslConstants.SelectAttribute);
            string selectValue = this.reader.Value;
            this.reader.MoveToAttribute(Constants.XslConstants.NameAttribute);
            variableName = PathCreation.RemovePrefixFromNode(this.reader.Value);
            if (this.IsScript(selectValue, false))
            {
                // if the variable is a function call to the CData section
                Controller.log.trace(MsgType.INFO, "A call is made by the variable to the CDATA section");
                value = selectValue.Substring(0, selectValue.IndexOf(':'));
                variableNode = InitializeVariableInitialization(null);
                selectValue = PathCreation.RemovePrefixFromNode(selectValue);
                string functionSource = selectValue.Substring(0, selectValue.IndexOf('('));
                FunctoidDictionary dictionary = new FunctoidDictionary();
                variableValue = Constants.TRFMConstants.VariableTag + TrfmCounters.VariableCounter;
                if (dictionary.IsSupported(functionSource) && !this.IsScript(value, true))
                {
                    // supported functoids
                    Controller.log.trace(MsgType.INFO, "The functoin call represents a supported functoid");
                    string functionTarget = dictionary.GetFunctoidName(functionSource);
                    FunctoidCreation functoidVar = new FunctoidCreation(functionTarget, selectValue, this.context);
                    object handledFunctoid = functoidVar.HandleConversion();
                    if (handledFunctoid == null)
                    {
                        TrfmCounters.VariableCounter--;
                        Controller.log.debug(MsgType.UNSUCCESSFUL_CONVERSION, "functoid " + functionSource.Substring(1) + " is supported by TRFM format but the conversion was unsucessfull");
                        return null;
                    }
                    else if (handledFunctoid is VariableInitialization)
                    {
                        // if a variable initialisation is returned
                        VariableInitialization temp = (VariableInitialization)handledFunctoid;

                        // update the current variable inialisation with the returned object properties
                        variableNode.Properties["XCell"] = temp.Properties["XCell"];
                        variableNode.Properties["YCell"] = temp.Properties["YCell"];
                        variableNode.Properties["Label"] =temp.Properties["Label"];
                        variableNode.Properties["Id"] = temp.Properties["Id"];
                        variableNode.Identifier = temp.Identifier;
                        variableValue = temp.Identifier;
                        handledFunctoid = temp.Initializer;
                    }

                    MapExpression me = (MapExpression)handledFunctoid;
                    if (me != null)
                    {
                        variableNode.Initializer = me;
                        if (functionTarget.Contains("Cumulative"))
                        {
                            return this.HandleCumulativeFunctoids(variableNode, variableName);
                        }
                    }
                    else
                    {
                        TrfmCounters.VariableCounter--;
                        Controller.log.debug(MsgType.UNSUCCESSFUL_CONVERSION, "Functoid conversion failed :" + selectValue);
                        return null;
                    }
                }
                else
                {
                    // all functoids that are not implemented  
                    // temporarily inserting an arithmetic expression functoid intead of a void functoid
                    Controller.log.trace(MsgType.INFO, "Unsupported function call");
                    if (this.UnconvertableFunctions(functionSource, selectValue))
                    {
                        return null;
                    }

                    if (this.ResolvePrefix(value) == CDataReaderConstants.LanguageType.CSharp)
                    {
                        Controller.log.trace(MsgType.SUCCESSFUL_CONVERSION, "CSharp Functoid-\"" + functionSource + '"');
                        cdataSection = this.GetCDataSection(functionSource, selectValue, value);
                        variableNode = this.AddFunctoid("CSharp", selectValue, false, cdataSection);
                        variableNode.Properties["Id"] = Constants.map1.Scripts.Last().ScriptId;
                    }
                    else
                    {
                        Controller.log.debug(MsgType.UNSUCCESSFUL_CONVERSION, "Script/Custom Functoid-\"" + functionSource + '"');
                        string funcTgt = "script/custom functoid-\"" + functionSource + '"';
                        variableNode = this.AddFunctoid(funcTgt, selectValue, true);
                        if (!dictionary.ContainsKey(functionSource))
                        {
                            cdataSection = this.GetCDataSection(functionSource, selectValue, value);
                            if (cdataSection.Length > 1024)
                            {
                                cdataSection = cdataSection.Substring(0, 1020) + "...";
                            }
                            variableNode.Properties.Add("Comments", cdataSection);
                        }
                    }

                    if (variableNode == null)
                    {
                        return null;
                    }
                    else
                    {
                        variableValue = variableNode.Identifier;
                    }
                }
            }
            else
            {
                // variables that do not denote functoids
                Controller.log.trace(MsgType.INFO, "The variable declared is parsed");
                object varObject = this.expressionHandler.ParseExpression(selectValue);
                variableValue = this.GetVariableValue(varObject, out variableNode);
                if (variableValue == null)
                {
                    return null;
                }
            }

            if (!string.IsNullOrEmpty(variableName))
            {
                Controller.log.trace(MsgType.SUCCESSFUL_CONVERSION, "Variable " + variableName + " generated");
                this.context.SetVarDictionary(variableName, variableValue);
            }
            else
            {
                Controller.log.debug(MsgType.UNSUCCESSFUL_CONVERSION, "Functoid not created");
                variableNode = null;
            }

            return variableNode;
        }

        internal static VariableInitialization InitializeVariableInitialization(string label = null)
        {
            if (string.IsNullOrEmpty(label))
            {
                label = string.Empty;
            }

            return new VariableInitialization
            {
                Properties = 
                {
                    { "XCell", Constants.TRFMConstants.XcellValue },
                    { "YCell", Constants.TRFMConstants.YcellValue },
                    { "Label", label },
                    { "Id", TrfmCounters.GetNewFunctoidId()}
                },
                Identifier = Constants.TRFMConstants.VariableTag + ++TrfmCounters.VariableCounter
            };
        }

        internal VariableInitialization AddFunctoid(string functoidName, string input, bool isAbsent, string scriptContent = null)
        {
            VariableInitialization temp = InitializeVariableInitialization();
            FunctoidDictionary dictionary = new FunctoidDictionary();
            string label = string.Empty;
            FunctoidCreation functoidVar;
            if (!isAbsent && dictionary.IsSupported(functoidName))
            {
                if (functoidName.StartsWith("Iteration"))
                {
                    ContextExpression contextExpression = new ContextExpression();
                    contextExpression.ContextFunctionName = ContextFunctionNames.GetCurrentPosition;
                    temp.Initializer = contextExpression;
                    return temp;
                }
                else if (functoidName.StartsWith("CSharp"))
                {
                    functoidVar = new FunctoidCreation(dictionary.GetFunctoidName(functoidName), input, this.context, scriptContent);
                }
                else
                {
                    functoidVar = new FunctoidCreation(dictionary.GetFunctoidName(functoidName), input, this.context);
                }
            }
            else
            {
                temp.Properties["Label"] = functoidName + " Not Converted Sucessfully ";
                Controller.log.debug(MsgType.UNSUCCESSFUL_CONVERSION, temp.Properties["Label"].ToString());
                functoidVar = new FunctoidCreation("2Notes", input, this.context);
            }

            MapExpression me = (MapExpression)functoidVar.HandleConversion();
            if (me == null)
            {
                return null;
            }

            temp.Initializer = me;
            return temp;
        }

        // in this function the input is the input value in trfm
        internal VariableInitialization CreateConstantFunctoid(string input, bool isConstantExpression)
        {
            VariableInitialization temp = InitializeVariableInitialization();
            FunctionInvocationExpression expression = new FunctionInvocationExpression();
            expression.FunctionName = "StringConcatenate";
            if (isConstantExpression)
            {
                expression.Parameters.Add(ExpressionArgumentCreator.CreateConstantExpression(input));
            }
            else
            {
                expression.Parameters.Add(ExpressionArgumentCreator.CreatePathExpression(input));
            }

            temp.Initializer = expression;
            return temp;
        }

        #region private functions

        // Determines the Variable value which matches to the corresponding variable declared in the BTM file
        private string GetVariableValue(object varObject, out VariableInitialization varibaleNode)
        {
            string varValue = string.Empty;
            varibaleNode = null;
            if (varObject == null)
            {
                return null;
            }
            else if (varObject.GetType() == typeof(ConstantExpression))
            {
                varValue = "\"" + ((ConstantExpression)varObject).Value.ToString() + "\"";
            }
            else if (varObject.GetType() == typeof(PathExpression))
            {
                varValue = ((PathExpression)varObject).Path.ToString();
            }
            else if (varObject.GetType() == typeof(VariableInitialization))
            {
                varibaleNode = (VariableInitialization)varObject;
                varValue = varibaleNode.Identifier;
            }
            else
            {
                Controller.log.debug(MsgType.ERROR, "Could not figure out what the select value denotes:" + this.reader.Value);
                return null;
            }

            return varValue;
        }

        // Checks if the function cannot be conveted to a functoid and does the required processing
        private bool UnconvertableFunctions(string functionSource, string selectValue)
        {
            if (this.IsInitCumulativeFunctoinCall(functionSource))
            {
                ListCreator listCreator = new ListCreator(this.context);
                this.context.CurrentListInitialisation = listCreator.ListInitializationNode;
                Controller.log.trace(MsgType.INFO, "IsInitCumulativeFunctoid found");
                return true;
            }
            else if (this.IsAddToCumulativeFunctoinCall(functionSource))
            {
                Controller.log.trace(MsgType.INFO, "IsAddToCumulativeFunctoid found");
                selectValue = selectValue.Substring(selectValue.IndexOf('(') + 1, selectValue.LastIndexOf(')') - selectValue.IndexOf('(') - 1);
                Constants.CumulativeInput = selectValue.Substring(selectValue.IndexOf('(') + 1, selectValue.LastIndexOf(')') - selectValue.IndexOf('(') - 1);

                // if the Add to cumulative happens within a list in trfm
                ListCreator listCreator = new ListCreator(this.context.CurrentListInitialisation, this.context);
                ITransformType listInputType;
                if (functionSource.Contains("Concat"))
                {
                    listInputType = SimpleType.String;
                }
                else
                {
                    listInputType = SimpleType.Number;
                }

                this.context.AddItemToCurrentContext(listCreator.AddItems(
                    new List<string>() { Constants.CumulativeInput },
                    new List<ITransformType>() { listInputType }));
                Controller.log.trace(MsgType.INFO, "input to cumulative functoid :" + Constants.CumulativeInput);

                return true;
            }
            else if (functionSource.StartsWith("DBLookupShutdown"))
            {
                Controller.log.trace(MsgType.INFO, "DBLookupShutdown function found");
                return true;
            }

            return false;
        }

        private object HandleCumulativeFunctoids(VariableInitialization handledFunctoid, string varName)
        {
            // adding a list in case it has been created
            if (this.context.CurrentListInitialisation != null)
            {
                while (this.UpdationRequiredInCurrentList()) 
                {
                }

                this.context.AddItemToCurrentContext(this.context.CurrentListInitialisation);
                this.context.CurrentListInitialisation = null;
            }

            FunctionInvocationExpression functoid = (FunctionInvocationExpression)handledFunctoid.Initializer;
            if (functoid.Target is PathExpression)
            {
                string path = ((PathExpression)functoid.Target).Path;
                functoid.Target = new ListPathExpression { Path = path };
            }

            if (functoid.Parameters.Count == 0)
            {
                SimpleType returnType;
                if (functoid.FunctionName == "CumulativeConcat")
                {
                    returnType = SimpleType.String;
                }
                else
                {
                    returnType = SimpleType.Number;
                }

                functoid.Parameters.Add(ExpressionArgumentCreator.CreateLambdaExpression("item", new List<string> { "item" }, returnType));
            }

            ListPathExpression param = (ListPathExpression)functoid.Target;
            if (!this.PresentInCurrentScope(param.Path, this.context))
            {
                if (param.Path.StartsWith(Constants.TRFMConstants.ScopeSourceHome))
                {
                    Constants.map1.Statements.Add(handledFunctoid);
                    Controller.log.trace(MsgType.SUCCESSFUL_CONVERSION, "Variable " + varName + " generated");
                }
                else
                {
                    this.context.AddItemToCurrentContext(handledFunctoid);
                }

                VariableInitialization temp = this.CreateConstantFunctoid(handledFunctoid.Identifier, false);
                this.context.SetVarDictionary(varName, temp.Identifier);
                return temp;
            }

            this.context.SetVarDictionary(varName, handledFunctoid.Identifier);
            return handledFunctoid;
        }

        // Updates the current list initialisation when add item has been added to the sae scope instead of inside the list
        private bool UpdationRequiredInCurrentList()
        {
            List<object> objectInContext = this.context.GetItemsInCurrentContext();
            int index = 0;
            foreach (object currentObject in objectInContext)
            {
                if (currentObject.GetType() == typeof(AddItemToList))
                {
                    this.context.CurrentListInitialisation.Initializer.Statements.Add((MapStatement)currentObject);
                    this.context.RemoveItemFromCurrentContext(index);
                    return true;
                }

                index++;
            }

            return false;
        }

        private bool PresentInCurrentScope(string path, OperationContext context)
        {
            string pathSource = string.Empty;
            if (path.StartsWith(Constants.TRFMConstants.VariableTag))
            {
                return true; // call list creator
            }

            if (path.IndexOf('/') >= 0)
            {
                pathSource = path.Substring(0, path.IndexOf('/'));
            }
            else
            {
                pathSource = path;
            }
            
            string contextTargetVariable = string.Empty;
            if (context.SubTreePath.IndexOf('/') > 0)
            {
                contextTargetVariable = context.SubTreePath.Substring(0, context.SubTreePath.IndexOf('/'));
            }
            else
            {
                contextTargetVariable = context.SubTreePath;
            }

            string contextSourceVariable = contextTargetVariable.Replace("tgt", "src");
            if (pathSource.Equals(contextSourceVariable))
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        // returns the CDATA function content using the function call
        private string GetCDataSection(string methodName, string inputs, string prefix)
        {
            string data = string.Empty;
            string[] array = { " , " };
            inputs = inputs.Substring(inputs.IndexOf('(') + 1);
            inputs = inputs.Substring(0, inputs.Length - 1).Trim();
            string[] input = inputs.Split(array, StringSplitOptions.RemoveEmptyEntries);
            MethodInfoProvider infoProvider = new MethodInfoProvider(DesignMapConvertor.XsltPath);
            data = infoProvider.GetMethodAndReferencedMethodsContent(methodName, input.Count(), this.ResolvePrefix(prefix));
            if (data == null)
            {
                data = string.Empty;
            }

            Controller.log.trace(MsgType.INFO, "CDATA Section read for a script functoid " + data);
            return data;
        }

        private CDataReaderConstants.LanguageType ResolvePrefix(string prefix)
        {
            if (prefix.Equals(Constants.XslConstants.CsharpFunctionPrefix))
            {
                return CDataReaderConstants.LanguageType.CSharp;
            }
            else if (prefix.Equals(Constants.XslConstants.JscriptFunctionPrefix))
            {
                return CDataReaderConstants.LanguageType.JavaScript;
            }
            else if (prefix.Equals(Constants.XslConstants.VbFunctionPrefix))
            {
                return CDataReaderConstants.LanguageType.VB;
            }
            else
            {
                return CDataReaderConstants.LanguageType.Unsupported;
            }
        }

        // Checks if the call is being made to the CDATA section
        // if Functoid check is true, then it checks if it is a script functoid(that is it excludes checks for csharp function calls and external access function calls)
        private bool IsScript(string value, bool functoidCheck)
        {
            if (value.Trim().StartsWith(Constants.XslConstants.VbFunctionPrefix))
            {
                return true;
            }
            else if (value.Trim().StartsWith(Constants.XslConstants.JscriptFunctionPrefix))
            {
                return true;
            }

            if (!functoidCheck)
            {
                if (value.Trim().StartsWith(Constants.XslConstants.CsharpFunctionPrefix))
                {
                    return true;
                }
                else if (value.Trim().StartsWith(Constants.XslConstants.ScriptNSFunctionPrefix))
                {
                    return true;
                }
            }

            return false;
        }

        private bool IsAddToCumulativeFunctoinCall(string functionName)
        {
            if (functionName.Equals("AddToCumulativeConcat"))
            {
                return true;
            }
            else if (functionName.Equals("AddToCumulativeAvg"))
            {
                return true;
            }
            else if (functionName.Equals("AddToCumulativeMax"))
            {
                return true;
            }
            else if (functionName.Equals("AddToCumulativeMin"))
            {
                return true;
            }
            else if (functionName.Equals("AddToCumulativeSum"))
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        private bool IsInitCumulativeFunctoinCall(string functionName)
        {
            if (functionName.Equals("InitCumulativeConcat"))
            {
                return true;
            }
            else if (functionName.Equals("InitCumulativeAvg"))
            {
                return true;
            }
            else if (functionName.Equals("InitCumulativeMax"))
            {
                return true;
            }
            else if (functionName.Equals("InitCumulativeMin"))
            {
                return true;
            }
            else if (functionName.Equals("InitCumulativeSum"))
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        #endregion
    }
}
